#

# Various I/O blocks.
#

import os
import param

import pandas as pd
import panel as pn

from sier2 import Block
from pathlib import Path
from io import StringIO, BytesIO

class LoadDataFrame(Block):
    """ GUI import from csv/excel file.

    """

    # Unfortunately, file selection in Panel is dodgy.
    # We need to use a FileInput widget, which uploads the file as a bytes object.
    #
    in_file = param.Bytes(label='Input File', doc='Bytes object of the input file.')
    in_header_row = param.Integer(label='Header row', default=0)
    out_df = param.DataFrame()

    def __init__(self, *args, block_pause_execution=True, **kwargs):
        super().__init__(*args, block_pause_execution=block_pause_execution, **kwargs)

        self.i_if = pn.widgets.FileInput.from_param(
            self.param.in_file,
            accept='.csv,.xlsx,.xls',
            multiple=False
        )

    def execute(self):
        pn.state.notifications.info('Reading file', duration=5_000)

        try:
            if self.i_if.filename.endswith('.csv'):
                self.out_df = pd.read_csv(StringIO(self.in_file.decode('utf-8')), header=self.in_header_row)
            elif self.i_if.filename.endswith('.xlsx') or self.i_if.filename.endswith('.xls'):
                self.out_df = pd.read_excel(BytesIO(self.in_file), header=self.in_header_row)

        except Exception as e:
            pn.state.notifications.error(f'Error reading {self.i_if.filename}. Check sidebar logs for more information.', duration=10_000)
            self.logger.error(str(e))
            #TODO: add feature to logger to send logs to Dag developer

    def __panel__(self):
        i_hr = pn.Param(
            self,
            parameters=['in_header_row'],
            widgets={
                'in_header_row': pn.widgets.IntInput
            }
        )
        return pn.Column(self.i_if, i_hr)

class SaveDataFrame(Block):
    """ Save a dataframe to a csv or xlsx.   
    """

    in_df = param.DataFrame()
    in_file_name = param.String()
    in_subset_len = param.Integer(label='Records to save', default=None)
    in_subset_type = param.String(label='Subset type', default='All')
    
    # Preferred to use default_filename=None and then set a default value like default_filename = '' if not default_filename else default_filename
    def __init__(self, *args, default_filename=None, **kwargs):
        super().__init__(*args, **kwargs)

        # We want to dynamically enable/disable buttons as the user sets the filename and subset info, 
        # so we declare widgets explicitly.

        # We should tell the user how big the file is going to be, just in case.
        #
        self.size_msg = pn.widgets.StaticText(
            value=''
        )

        self.i_fn = pn.widgets.TextInput.from_param(
            self.param.in_file_name,
            placeholder='Output file name',
            value='' if not default_filename else default_filename,
            name='Output file name (without extension)'
        )

        self.i_sub_l = pn.widgets.IntInput.from_param(
            self.param.in_subset_len, 
            value=100, 
            step=10, 
            start=0,
            disabled=True,
        )

        self.i_sub_t = pn.widgets.ToggleGroup.from_param(
            self.param.in_subset_type,
            options=['All', 'Head', 'Tail', 'Random sample'], 
            behavior="radio",
        )
        
        self.csvdl = pn.widgets.FileDownload(
            callback=self.download_csv,
            button_type='success',
            filename='',
            label='Download .csv'
        )

        self.xlsxdl = pn.widgets.FileDownload(
            callback=self.download_xlsx,
            button_type='success',
            filename='',
            label='Download .xlsx'
        )

        self.csvdl.disabled = True
        self.xlsxdl.disabled = True

        # Hook up the filename widget to the download widget.
        # Make sure to watch value_input, which is updated live as the user edits the TextInput.
        # 'value' is only updated if the user hits enter.
        #
        def update_name(event):
            if self.i_fn.value_input:
                self.xlsxdl.disabled = False
                self.xlsxdl.filename = f'{self.i_fn.value_input}.xlsx'
                self.csvdl.disabled = False
                self.csvdl.filename = f'{self.i_fn.value_input}.csv'
            else:
                self.xlsxdl.disabled = True
                self.csvdl.disabled = True

        self.i_fn.param.watch(update_name, 'value_input')

        # Make sure we disable the subset length selection if we're saving everything.
        def update_subset(event):
            if self.i_sub_t.value == 'All':
                self.i_sub_l.value = self.in_df.shape[0]
                self.i_sub_l.disabled = True
            else:
                self.i_sub_l.value = 100
                self.i_sub_l.disabled = False

        self.i_sub_t.param.watch(update_subset, 'value')

    def download_csv(self):
        buf = StringIO()
        
        if self.in_subset_type == 'All':
            self.in_df.to_csv(buf)
        elif self.in_subset_type == 'Head':
            self.in_df.head(self.in_subset_len).to_csv(buf)
        elif self.in_subset_type == 'Tail':
            self.in_df.tail(self.in_subset_len).to_csv(buf)
        elif self.in_subset_type == 'Random sample':
            self.in_df.sample(self.in_subset_len).to_csv(buf)
        else:
            # We shouldn't ever end up here.
            raise NotImplementedError
        
        buf.seek(0)
        return buf

    def download_xlsx(self):
        buf = BytesIO()
        writer = pd.ExcelWriter(buf, engine='xlsxwriter')
        
        if self.in_subset_type == 'All':
            self.in_df.to_excel(writer)
        elif self.in_subset_type == 'Head':
            self.in_df.head(self.in_subset_len).to_excel(writer)
        elif self.in_subset_type == 'Tail':
            self.in_df.tail(self.in_subset_len).to_excel(writer)
        elif self.in_subset_type == 'Random sample':
            self.in_df.sample(self.in_subset_len).to_excel(writer)
        else:
            # We shouldn't ever end up here.
            raise NotImplementedError
            
        writer.close()
        buf.seek(0)
        return buf

    def execute(self):
        self.size_msg.value = f'Saving data frame of size {self.in_df.shape}.'

        # Only allow file download if we've set an input.
        #
        if self.i_fn.value_input:
            self.csvdl.disabled = False
            self.xlsxdl.disabled = False

        self.i_sub_l.value = self.in_df.shape[0]
        
    def __panel__(self):
        return pn.Column(
            self.size_msg,
            self.i_fn,
            self.i_sub_t,
            self.i_sub_l,
            pn.Row(self.csvdl, self.xlsxdl),
        )




import param
import pandas as pd
import panel as pn
from sier2 import Block

class SimpleTable(Block):
    """ Simple Table Viewer

    Make a tabulator to display an input table.
    """

    in_df = param.DataFrame(doc='Input pandas dataframe')
    out_df = param.DataFrame(doc='Output pandas dataframe', default=pd.DataFrame())

    def execute(self):
        self.out_df = self.in_df

    def __panel__(self):
        return pn.Param(
            self,
            parameters=['out_df'],
            widgets={'out_df': {'type': pn.widgets.Tabulator, 'page_size':20, 'pagination':'local', 'name':'DataFrame'}}
        )

class SimpleTableSelect(Block):
    """ Simple Table Selection

    Make a tabulator to display an input table.
    Pass on selections as an output.
    """

    in_df = param.DataFrame(doc='Input pandas dataframe')
    out_df = param.DataFrame(doc='Output pandas dataframe')

    def __init__(self, *args, block_pause_execution=True, **kwargs):
        super().__init__(*args, block_pause_execution=block_pause_execution, continue_label='Continue With Selection', **kwargs)
        self.tabulator = pn.widgets.Tabulator(pd.DataFrame(), name='DataFrame', page_size=20, pagination='local')

    # def prepare(self):
    #     if self.in_df is not None:
    #         self.tabulator.value = self.in_df
    #     else:
    #         self.tabulator.value = pd.DataFrame()

    def execute(self):
        self.out_df = self.tabulator.selected_dataframe

    def __panel__(self):
        return self.tabulator


from ..blocks.io import LoadDataFrame, SaveDataFrame
from ..blocks.view import SimpleTable, SimpleTableSelect
from ..blocks.holoviews import HvPoints, HvPointsSelect, HvHist
from ..blocks.test_data import StaticDataFrame, FakerData

from sier2 import Connection
from sier2.panel import PanelDag


def hv_points():
    """Load a dataframe from a file and display a Points chart."""

    ldf = LoadDataFrame(name='Load DataFrame')
    hps = HvPointsSelect(name='Plot Points')
    st = SimpleTable(name='View Selection')

    DOC = '''# Points chart
    
    Load a dataframe from a file and display a Points chart.
    '''

    dag = PanelDag(doc=DOC, site='Chart', title='Points')
    dag.connect(ldf, hps,
        Connection('out_df', 'in_df'),
    )
    dag.connect(hps, st,
        Connection('out_df', 'in_df'),
    )

    return dag

def hv_hist():
    """Load a dataframe from a file and display a Histogram."""

    ldf = LoadDataFrame(name='Load DataFrame')
    hh = HvHist(name='Plot Histogram')

    DOC = '''# Points chart
    
    Load a dataframe from a file and display a Points chart.
    '''

    dag = PanelDag(doc=DOC, site='Chart', title='Histogram')
    dag.connect(ldf, hh,
        Connection('out_df', 'in_df'),
    )

    return dag

def table_view():
    """Load a dataframe from file and display in a panel table."""

    ldf = LoadDataFrame(name='Load DataFrame')
    st = SimpleTableSelect(name='View Table')
    sel_st = SimpleTable(name='Selection')

    DOC = '''# Table viewer

    Load a dataframe from a file and display the data as a table.
    '''

    dag = PanelDag(doc=DOC, title='Table')
    dag.connect(ldf, st, Connection('out_df', 'in_df'))
    dag.connect(st, sel_st, Connection('out_df', 'in_df'))

    return dag

def faker_view():
    """Load and display fake data."""

    fdf = FakerData(name='Fake Data')
    st = SimpleTable(name='Display')

    DOC = '''# Faker example

    Load and display fake data.
    '''

    dag = PanelDag(doc=DOC, title='Faker')
    dag.connect(fdf, st, Connection('out_data', 'in_df'))

    return dag

def static_view():
    """Load a static example dataframe and display in a table."""

    sdf = StaticDataFrame(name='Load DataFrame')
    st = SimpleTableSelect(name='View Table')
    sel_st = SimpleTable(name='Selection')

    DOC = '''# Static table viewer

    Load a static example dataframe and display the data as a table.
    '''

    dag = PanelDag(doc=DOC, title='Table')
    dag.connect(sdf, st, Connection('out_df', 'in_df'))
    dag.connect(st, sel_st, Connection('out_df', 'in_df'))

    return dag

def save_csv():
    """Load a dataframe from file and download."""
    sdf = StaticDataFrame(name='Load DataFrame')
    st = SimpleTableSelect(name='View Table')
    edf = SaveDataFrame(name='Export')

    DOC = '''# Table viewer

    Load a dataframe from file and download.
    '''

    dag = PanelDag(doc=DOC, title='Table')
    dag.connect(sdf, st, Connection('out_df', 'in_df'))
    dag.connect(st, edf, Connection('out_df', 'in_df'))

    return dag


from sier2 import Info

def blocks() -> list[Info]:
    return [
        Info('sier2_blocks.blocks.io:LoadDataFrame', 'Load a dataframe from a file'),
        Info('sier2_blocks.blocks.io:SaveDataFrame', 'Save a dataframe'),
        
        Info('sier2_blocks.blocks.view:SimpleTable', 'Display a simple table'),
        Info('sier2_blocks.blocks.view:SimpleTableSelect', 'Display a simple table and pass selections on'),

        Info('sier2_blocks.blocks.holoviews:HvPoints', 'A Holoviews Points chart'),
        Info('sier2_blocks.blocks.holoviews:HvPointsSelect', 'A Holoviews Points chart that passes on selections'),
        Info('sier2_blocks.blocks.holoviews:HvHist', 'A Holoviews Histogram chart'),

        Info('sier2_blocks.blocks.test_data:StaticDataFrame', 'Static test dataframe'),
        Info('sier2_blocks.blocks.test_data:FakerData', 'Generate realistic fake data of various types'),
    ]

def dags() -> list[Info]:
    return [
        Info('sier2_blocks.dags.examples:table_view', 'Load a dataframe from file and display in a panel table'),
        Info('sier2_blocks.dags.examples:static_view', 'Load a static dataframe and display in a panel table'),
        Info('sier2_blocks.dags.examples:save_csv', 'Load and export a dataframe'),
        Info('sier2_blocks.dags.examples:hv_points', 'Load and plot a dataframe as points'),
        Info('sier2_blocks.dags.examples:hv_hist', 'Load a dataframe and plot a histogram'),
        Info('sier2_blocks.dags.examples:faker_view', 'Load and display fake data'),
    ]



from sier2 import Block
import param

import panel as pn
import numpy as np

import holoviews as hv
hv.extension('bokeh', inline=True)

## TODO: Option to force plots to have equal axes sizes on initialisation.
## ^^ Not sure this is possible nicely within Bokeh. You can set data_aspect, but that
## has awkward effects with panel's stretch sizing.

class HvPoints(Block):
    """The Points element visualizes as markers placed in a space of two independent variables."""

    in_df = param.DataFrame(doc='A pandas dataframe containing x,y values')
    out_df = param.DataFrame(doc='Output pandas dataframe')
    
    x_sel = param.ObjectSelector()
    y_sel = param.ObjectSelector()
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.hv_pane = pn.pane.HoloViews(sizing_mode='stretch_width', min_height=600)
        self.hv_pane.object=self._produce_plot

    @param.depends('in_df', 'x_sel', 'y_sel')
    def _produce_plot(self):
        if None not in (self.x_sel, self.y_sel):
            return hv.Points(self.in_df, kdims=[self.x_sel, self.y_sel])

        else:
            return hv.Points([])

    def execute(self):
        plottable_cols = [c for c in self.in_df.columns if self.in_df[c].dtype.kind in 'iuf']
        if len(plottable_cols)<2:
            pn.state.notifications.error(f'Error plotting. Could not find two columns with numeric data.', duration=10_000)
        else:
            self.param['x_sel'].objects = plottable_cols
            self.param['y_sel'].objects = plottable_cols
            self.x_sel = plottable_cols[0]
            self.y_sel = plottable_cols[1]

        self.out_df = self.in_df

    def __panel__(self):
        return pn.Column(
            pn.Row(
                self.param['x_sel'],
                self.param['y_sel']
            ),
            self.hv_pane
        )

class HvPointsSelect(Block):
    """The Points element visualizes as markers placed in a space of two independent variables."""

    in_df = param.DataFrame(doc='A pandas dataframe containing x,y values')
    out_df = param.DataFrame(doc='Output pandas dataframe')

    def __init__(self, block_pause_execution=True, *args, **kwargs):
        super().__init__(
            *args,
            block_pause_execution=block_pause_execution,
            continue_label='Continue With Selection',
            **kwargs,
        )
        
        self.hv_pane = pn.pane.HoloViews(sizing_mode='stretch_width', min_height=600)
        self.selection = hv.streams.Selection1D()
        self.hv_pane.object=self._produce_plot

    x_sel = param.ObjectSelector()
    y_sel = param.ObjectSelector()
    
    @param.depends('in_df', 'x_sel', 'y_sel')
    def _produce_plot(self):
        if self.in_df is not None and self.x_sel is not None and self.y_sel is not None:
            scatter = hv.Points(self.in_df, kdims=[self.x_sel, self.y_sel])

        else:
            scatter = hv.Points([])

        scatter = scatter.opts(tools=['box_select'])
        self.selection.source = scatter
        return scatter

    def prepare(self):
        plottable_cols = [c for c in self.in_df.columns if self.in_df[c].dtype.kind in 'iuf']
        if len(plottable_cols)<2:
            pn.state.notifications.error(f'Error plotting. Could not find two columns with numeric data.', duration=10_000)
        else:
            self.param['x_sel'].objects = plottable_cols
            self.param['y_sel'].objects = plottable_cols
            self.x_sel = plottable_cols[0]
            self.y_sel = plottable_cols[1]

    def execute(self):
        self.out_df = self.in_df.loc[self.selection.index]

    def __panel__(self):
        return pn.Column(
            pn.Row(
                self.param['x_sel'],
                self.param['y_sel']
            ),
            self.hv_pane
        )

class HvHist(Block):
    """Produce a Histogram of input data."""

    # Define params
    #
    in_df = param.DataFrame(doc='A pandas dataframe containing x,y values')
    out_df = param.DataFrame(doc='Output pandas dataframe')

    column = param.ObjectSelector()
    bins = param.Integer(100, bounds=(0, None))
    integer_bins = param.Boolean(False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Make a pane to hold the plot.
        #
        self.hv_pane = pn.pane.HoloViews(sizing_mode='stretch_width', min_height=600)

        # Make a widget to control the bins parameter.
        # We need this so we can disable it when using integer bins.
        #
        self.bins_widget = pn.widgets.IntInput(name='Bins', start=0, value=self.bins)
        self.bins_widget.param.watch(lambda event: setattr(self, 'bins', event.new), 'value')

        # Set up the plot.
        #
        self.hv_pane.object=self._produce_plot
        
    
    @param.depends('in_df', 'column', 'bins', 'integer_bins')
    def _produce_plot(self):     

        # Disable the bin selection when using integer bins.
        #
        if self.integer_bins:
            self.bins_widget.disabled = True
        else:
            self.bins_widget.disabled = False
          
        if self.in_df is not None and self.column is not None :
            if self.integer_bins:
                col = self.in_df[self.column]
                self.bins = col.max() - col.min() + 1
                self.bins_widget.value = self.bins
                
                frequencies, edges = np.histogram(
                    col, 
                    np.arange(col.min()-0.5, col.max()+1.5, 1),
                )
            else:
                frequencies, edges = np.histogram(self.in_df[self.column], self.bins)

        else:
            frequencies, edges = [], []

        return hv.Histogram((edges, frequencies))        

    def execute(self):
        plottable_cols = [c for c in self.in_df.columns if self.in_df[c].dtype.kind in 'iuf']
        
        self.param['column'].objects = plottable_cols
        self.column = plottable_cols[0]

        self.out_df = self.in_df

    def __panel__(self):
        return pn.Column(
            pn.Row(
                self.param['column'],
                self.bins_widget,
                self.param['integer_bins'],
            ),
            self.hv_pane
        )


#

# Various blocks that produce input data, useful for testing dags.
#

import param

import pandas as pd
import panel as pn

from sier2 import Block

from faker import Faker


class StaticDataFrame(Block):
    """ Import static data frame for testing dags.

    """

    out_df = param.DataFrame()

    def execute(self):
        self.out_df = pd.DataFrame(data = {
            "calories": [420, 380, 390, None],
            "duration": [50, 40, 45, None],
            "Latitude": [0, 45, 70, None],
            "Longitude": [15, 30, 60, None],
            "Name": ['a', 'b', 'c', None],
        })

class FakerData(Block):
    """ Generate realistic fake data of various types.
    
    """

    in_data_type = param.ObjectSelector(default='name')
    in_output_type = param.ObjectSelector(
        default='dataframe', 
        objects=['dataframe', 'list']
    )
    in_output_length = param.Integer(default=100)
    
    out_data = param.ClassSelector(class_=(pd.DataFrame, list))

    def __init__(self, block_pause_execution=True, *args, **kwargs):
        super().__init__(
            block_pause_execution=block_pause_execution,
            continue_label='Generate fake data',
            *args, **kwargs,
        )
        self.fake = Faker()

        # The fake object can generate all sorts of things for us.
        # We can get a list of all the options from the attributes.
        #
        self.param['in_data_type'].objects = [
            opt for opt in dir(self.fake) if not opt.startswith('_') and not opt.endswith('_')
        ]

    def execute(self):
        gen = getattr(self.fake, self.in_data_type)
        data = [gen() for _ in range(self.in_output_length)]

        if self.in_output_type == 'dataframe':
            self.out_data = pd.DataFrame({self.in_data_type: data})

        else:
            self.out_data = data




